/* DDPref.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * Created on Oct 10, 2005
 * @author Eric Eaton
 */
package ddpref;

import weka.core.Instances;
import weka.core.Instance;
import weka.core.InstanceComparator;
import weka.core.Utils;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Random;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.HashMap;


import ddpref.data.BlocksGenerator;

public class DDPref {
	
	private Instances m_data;
	
	
	private boolean m_verbose = false;
	public void setVerbose(boolean v) { m_verbose = v; }
	public boolean getVerbose() { return m_verbose; }
	
	
	private Preferences m_preferences;
	public void setPreferences(Preferences prefs) {
		m_preferences = prefs;
	}
	public Preferences getPreferences() {
		return m_preferences;
	}

	
	
	double m_objFnValue = Double.NaN;
	public double getObjFnValue() { return m_objFnValue;}

	
	public DDPref() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	   * Returns a string describing this classifier
	   * @return a description of the classifier suitable for
	   * displaying in the explorer/experimenter gui
	   */
	public String globalInfo() {
		return "Class for building and using the DD-PREF Preference Algorithm."
	      +"For more information, see \n\n"
	      +"Marie desJardins and Kiri L. Wagstaff. "
	      +"DD-PREF: A Language for Expressing Preferences over Sets. "
	      +"AAAI 2005.";
	}
	
	
	  private Random m_randGenerator = null;
	  private long m_randomSeed = -1;
	  public void setRandomSeed(long seed) {
		  if (seed < 0) {
			  throw new IllegalArgumentException("Seed must be positive.");
		  }
		  m_randomSeed = seed;
	  }
	  public long getRandomSeed() { return m_randomSeed; }
	  
	  private void initRandomGenerator() {
		  // use the current time if no seed specified
		  if (m_randomSeed < 0) {
			  m_randGenerator = new Random(System.currentTimeMillis());
		  }
		  // otherwise use the specified seed
		  else {
			  m_randGenerator = new Random(m_randomSeed);
		  }
	  }

	  
	  /**
	   * Generates the classifier.
	   *
	   * @param instances set of instances serving as training data 
	   * @exception Exception if the classifier has not been generated successfully
	   */
	public void setData(Instances data) throws Exception {
		// TODO Auto-generated method stub
		// initialize the random number generator
		initRandomGenerator();
		
		// store the data
		m_data = data;
	}
	
	
	/** Performs subset selection using the specified method
	 * @param method The desired selection method.
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances selectSubset(SelectionMethod method, int k) {
		if (method == SelectionMethod.exhaustiveSelectSubset)
			return exhaustiveSelectSubset(k);
		else if (method == SelectionMethod.greedySelectSubset)
			return greedySelectSubset(k);
		else if (method == SelectionMethod.randomSelectSubset)
			return randomSelectSubset(k);
		else if (method == SelectionMethod.wrapperSelectSubset)
			return wrapperSelectSubset(k);
		else
			throw new IllegalArgumentException("Improper method specified");
		
	}
	
	/** Performs subset selection using the wrapper method.
	 * 
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances wrapperSelectSubset(int k) {
		if (m_preferences == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		return wrapperSelectSubset(k, m_preferences);
	}
	
	/** Performs subset selection using the wrapper method.
	 * 
	 * @param k The desired size of the subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances wrapperSelectSubset(int k, Preferences prefs) {
		return wrapperSelectSubset(k,null,prefs);
	}
	
	private static InstanceComparator m_comparator = new InstanceComparator();

	/** Performs subset selection using the wrapper method.
	 * 
	 * @param k The desired size of the subset.
	 * @param seedInstances Instances to seed the selected subset; these 
	 *                      instances will be included in the selected subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	private Instances wrapperSelectSubset(int k, Instances seedInstances, Preferences prefs) {
		m_objFnValue = Double.NaN;
		
		// temp vars to hold best selection so far
		Instances selectedInstances = null;
		Instances tempSelectedInstances = null;
		double maxObjFnValue = Double.MIN_VALUE;
		
		int sizeData = m_data.numInstances();
 		
		// loop through each item as the seed value to greedy selection
		for (int i=0; i<sizeData; i++) {
			
			// add the starting item to the seedSubset
			Instances seedSet = new Instances(m_data,0);
			Instance seedInstance = m_data.instance(i);
			seedSet.add(seedInstance);
			
			// initialize the seedSet from the given seed instances
			if (seedInstances != null) {
				for (int instIdx=0; instIdx<seedInstances.numInstances(); instIdx++) {
					Instance tempInstance = seedInstances.instance(instIdx);
					// add the instance only if it's different from the seed instance
					if (m_comparator.compare(seedInstance, tempInstance) != 0)
						seedSet.add(seedInstances.instance(instIdx));
				}
			}
			
			// extend the seedSet by greedy selection
			tempSelectedInstances = greedySelectSubset(k,seedSet, prefs);
			
			// keep the best seedSubset so far
			if (m_objFnValue > maxObjFnValue) {
				maxObjFnValue = m_objFnValue;
				selectedInstances = tempSelectedInstances;
			}
		}
		
		// set the final obj fn value
		m_objFnValue = maxObjFnValue;
		
		// return the selected instances
		return selectedInstances;
	}
	

	
	/** Selects a random subset.
	 * 
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances randomSelectSubset(int k) {
		if (m_preferences == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		return randomSelectSubset(k,m_preferences);
	}
	
	/** Selects a random subset.
	 * 
	 * @param k The desired size of the subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances randomSelectSubset(int k, Preferences prefs) {
		return randomSelectSubset(k,null,prefs);
	}
	
	/** Selects a random subset.
	 * 
	 * @param k The desired size of the subset.
	 * @param seedInstances Instances to seed the selected subset; these 
	 *                      instances will be included in the selected subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances randomSelectSubset(int k, Instances seedInstances, Preferences prefs) {
		
		// check for all preconditions satisfied
		if (m_data == null) {
			throw new IllegalStateException("Dataset not defined");
		}
		
		// check size of seedInstances
		if (seedInstances != null && seedInstances.numInstances() >= k) {
			System.err.println("Warning:  seedInstances already contains more items than needed.  Returning seedInstances.");
			return seedInstances;
		}
		
		// check that k isn't bigger than the dataset
		if (k > m_data.numInstances()) {
			System.err.println("Warning:  not enough data items for RandomSelectSubset():  requested "+k+" items, data only has "+m_data.numInstances()+" items.");
			m_objFnValue = prefs==null ? Double.NaN : computeObjFunction(m_data, prefs);
			return m_data;
		}
		
		// TreeSet<Instance> subset = new TreeSet<Instance>(m_comparator);
		HashSet<Instance> subset = new HashSet<Instance>();
		
		// initialize the seedSet from the given seed instances
		if (seedInstances != null) {
			for (int instIdx=0; instIdx<seedInstances.numInstances(); instIdx++) {
				subset.add(seedInstances.instance(instIdx));
			}
		}

		// select random items until the subset is of size k
		while (subset.size()<k)
			subset.add(m_data.instance(m_randGenerator.nextInt(m_data.numInstances())));
		
		// extract the items chosen
		Instances selectedInstances = collectionToInstances(subset);
		
		// compute the objective function value and return the subset
		m_objFnValue = prefs==null ? Double.NaN : computeObjFunction(selectedInstances, prefs);
		return selectedInstances;
	}
	
	
	/** Performs subset selection using exhaustive search.
	 * 
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances exhaustiveSelectSubset(int k) {
		if (m_preferences == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		return exhaustiveSelectSubset(k,m_preferences);
	}
	
	/** Performs subset selection using exhaustive search.
	 * 
	 * @param k The desired size of the subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances exhaustiveSelectSubset(int k, Preferences prefs) {

		// check for all preconditions satisfied
		if (m_data == null) {
			throw new IllegalStateException("Dataset not defined");
		}
		if (prefs == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		
		// check that k isn't bigger than the dataset
		if (k > m_data.numInstances()) {
			System.err.println("Warning:  not enough data items for ExhaustiveSelectSubset():  requested "+k+" items, data only has "+m_data.numInstances()+" items.");
			m_objFnValue = prefs==null ? Double.NaN : computeObjFunction(m_data, prefs);
			return m_data;
		}
		
		// loop through every subset of the indices, tracking the one with the max obj function value
		double maxObjFnValue = Double.MIN_VALUE;
		double objFnValue;
		Instances bestSubset = null;
		TreeSet<TreeSet<Instance>> subsets = getAllSubsets(m_data,k);
		for (TreeSet<Instance> subset : subsets) {
			Instances tempSubset = collectionToInstances(subset);
			objFnValue = computeObjFunction(tempSubset, prefs);
			if (objFnValue > maxObjFnValue) {
				maxObjFnValue = objFnValue;
				bestSubset = tempSubset;
			}
		}
		
		m_objFnValue = maxObjFnValue;
		return bestSubset;
	}
	
	private TreeSet<TreeSet<Instance>> getAllSubsets(Instances set, int k) {
		TreeSet<TreeSet<Instance>> subsets = new TreeSet<TreeSet<Instance>>( new Comparator<TreeSet<Instance>>() {
			public int compare(TreeSet<Instance> a, TreeSet<Instance> b) {
				Iterator<Instance> iterA = a.iterator();
				Iterator<Instance> iterB = b.iterator();
				InstanceComparator comparator = new InstanceComparator();
				// compare every item
				while (iterA.hasNext() && iterB.hasNext()) {
					Instance iA = iterA.next();
					Instance iB = iterB.next();
					int comparison = comparator.compare(iA,iB);
					// if they differ, determine based on the differing elements
					if (comparison != 0) {
						return comparison;
					}
				}
				// if they both ran out at the same time and have been the same up to this point, they're equal
				if (!iterA.hasNext() && !iterB.hasNext()) return 0;
				// if a ran out before b, then b is greater
				if (!iterA.hasNext()) return -1;
				// if b ran out before a, then a is greater
				if (!iterB.hasNext()) return 1;
				return 0; // dummy call -- will never be reached
			}
		});
		if (k<0) throw new IllegalArgumentException("k must be a positive integer.");
		if (k==0)
			return subsets;
		// if k=1, generate all subsets of size 1 directly from set
		if (k==1) {
			//generate all subsets of size 1 directly from set
			for (int instIdx=0; instIdx<set.numInstances(); instIdx++) {
				TreeSet<Instance> newSet = new TreeSet<Instance>(m_comparator);
				newSet.add(set.instance(instIdx));
				subsets.add(newSet);
			}
			return subsets;
		}
		// get the set of all subsets of size k-1
		TreeSet<TreeSet<Instance>> subsubsets = getAllSubsets(set,k-1);
		// generate each subset of size k from the subsets of size k-1
		for (Set<Instance> s : subsubsets) {
			// add each item to each subset of size k-1
			for (int instIdx=0; instIdx<set.numInstances(); instIdx++) {
				TreeSet<Instance> newSet = new TreeSet<Instance>(m_comparator);
				newSet.addAll(s);
				newSet.add(set.instance(instIdx));
				// only save it if the size is correct
				if (newSet.size() == k) subsets.add(newSet);
			}
		}
		return subsets;
	}
	
	/** Selects the top items by depth.
	 * 
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances topSelectSubset(int k) {
		if (m_preferences == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		return topSelectSubset(k,m_preferences);
	}
	
	/** Selects the top items by depth.
	 * 
	 * @param k The desired size of the subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances topSelectSubset(int k, Preferences prefs) {
		
		// check for all preconditions satisfied
		if (m_data == null) {
			throw new IllegalStateException("Dataset not defined");
		}
		if (prefs == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		
		// check that k isn't bigger than the dataset
		if (k > m_data.numInstances()) {
			System.err.println("Warning:  not enough data items for TopSelectSubset():  requested "+k+" items, data only has "+m_data.numInstances()+" items.");
			m_objFnValue = prefs==null ? Double.NaN : computeObjFunction(m_data, prefs);
			return m_data;
		}
		
		// loop through every item and record its depth
		double[] depths = new double[m_data.numInstances()];
		for (int instIdx=0; instIdx<m_data.numInstances(); instIdx++) {
			Instances tempInstances = new Instances(m_data,0);
			tempInstances.add(m_data.instance(instIdx));
			depths[instIdx] = depth(tempInstances,prefs);			
		}
		
		Instances selectedInstances = new Instances(m_data,0);
		
		// sort the depths, and keep the top most items
		double[] sortedDepths = new double[depths.length];
		System.arraycopy(depths,0,sortedDepths,0,depths.length);
		Arrays.sort(sortedDepths);
		double[] topDepths = new double[k - selectedInstances.numInstances()];
		System.arraycopy(sortedDepths,sortedDepths.length-topDepths.length,topDepths,0,topDepths.length);
		
		// loop through the unsorted depths, and pick out items with the top depths
		for (int instIdx=0; instIdx<m_data.numInstances(); instIdx++) {
			for (int i=0; i<topDepths.length; i++) {
				if (depths[instIdx] == topDepths[i]) {
					topDepths[i] = Double.MIN_VALUE;  // set it so that this depth can't ever be matched again
					selectedInstances.add(m_data.instance(instIdx));  // add the instance with that depth
					break; // break out of the inner loop
				} else if (depths[instIdx] < topDepths[i]) {
					break; // break, since topDepths is stored in increasing order, so the depth will never be match a later item
				}
			}
		}
		
		m_objFnValue = computeObjFunction(selectedInstances, prefs);
		return selectedInstances;
	}
	
	/** Performs subset selection using greedy selection.
	 * 
	 * @param k The desired size of the subset.
	 * @return The selected subset.
	 */
	public Instances greedySelectSubset(int k) {
		if (m_preferences == null) {
			throw new IllegalStateException("Preferences not defined.");
		}
		return greedySelectSubset(k,null,m_preferences);
	}
	
	/** Performs subset selection using greedy selection.
	 * 
	 * @param k The desired size of the subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	public Instances greedySelectSubset(int k, Preferences prefs) {
		return greedySelectSubset(k,null,prefs);
	}
		

	/** Performs subset selection using greedy selection.
	 * 
	 * @param k The desired size of the subset.
	 * @param seedInstances Instances to seed the selected subset; these 
	 *                      instances will be included in the selected subset.
	 * @param prefs The preferences used in the selection.
	 * @return The selected subset.
	 */
	private Instances greedySelectSubset(int k, Instances seedInstances, Preferences prefs) {
		// check for all preconditions satisfied
		if (m_data == null) {
			throw new IllegalStateException("Dataset not defined");
		}
		if (prefs == null) {
			throw new IllegalStateException("Preferences not defined.");
		}

		// error check size of seedSubset
		if (seedInstances != null && seedInstances.numInstances() > k) {
			throw new IllegalArgumentException("Size of seedInstances cannot be greater than k. k="+k+" and |seedInstances|="+seedInstances.numInstances());
		}
		
		// initialize the subset and objective function
		//Set<Instance> subset = new HashSet<Instance>(m_comparator);
		Set<Instance> subset = new HashSet<Instance>();
		if (seedInstances != null && seedInstances.numInstances() > 0) {
			for (int instIdx=0; instIdx<seedInstances.numInstances(); instIdx++) {
				subset.add(seedInstances.instance(instIdx));
			}
		}
		double objFnValue = Double.NaN;
		
		// gather the k element subset
		while (subset.size() < k) {
			try {
				objFnValue = greedyAddItem(subset, prefs);
			} catch (IllegalStateException e) {
				System.err.println(e.getMessage());
				System.err.println("Warning: not enough satisfying items in greedySelectSubset():  "
						+"asked for "+k+" items, but only returning "+subset.size()+" items.");
				m_objFnValue = objFnValue;
				// convert and return the subset
				return collectionToInstances(subset);
			}
		}
		
		m_objFnValue = objFnValue;
		
		// convert and return the subset
		return collectionToInstances(subset);
	}
	
	/** Helper function to convert a collection of instances to weka's Instances
	 * @param collection The collection to convert.
	 * @return The collection as a weka Instances.
	 */
	private Instances collectionToInstances(Collection<Instance> collection) {
		Instances instances = new Instances(m_data,0);
		for (Instance inst : collection) {
			instances.add(inst);
		}
		return instances;
	}
	
	
	/** Adds an item to the given subset, using greedy selection.
	 * @param subset The given subset.
	 * @return The objective function value for the modified subset.
	 * @throws IllegalStateException
	 */
	private double greedyAddItem(Set<Instance> subset, Preferences prefs) throws IllegalStateException{
		double maxObjFnValue = Double.NEGATIVE_INFINITY;
		
		// list to hold all candidate items which yield the same highest obj fn value
		ArrayList<Instance> selectedItemsList = new ArrayList<Instance>();
		
		/*new1 */
		Instances subsetInstances = collectionToInstances(subset);
		
		// loop through every data instance
		int numInstances = m_data.numInstances();
		
		//new8 compute the depth of the set without adding anything
		double baseDepth = depth(subsetInstances, prefs, false, 0, 0, true);
		int    baseSize  = subsetInstances.numInstances();
		
		for (int instIdx = 0; instIdx<numInstances; instIdx++) {
			
			Instance candidateInstance = m_data.instance(instIdx);
			
			// if the subset doesn't contain this instance, evaluate adding it
			if (!subset.contains(candidateInstance)) {
				
				// convert the subset to a set of instances
				/*old1 Instances subsetInstances = collectionToInstances(subset); why compute every time*/
				
				// compute the value of the objective function for adding this instance
				subsetInstances.add(candidateInstance);  // adds to last index
				
				/*old2 double value = computeObjFunction(subsetInstances, prefs); made inline bellow*/
				/*new2 inline of above*/
				/*old5 
				 double value += (  (prefs.alpha() * diversityMatch(subsetInstances, prefs, true) 
						           + (1.0-prefs.alpha()) * depth(subsetInstances, prefs, true))  )
				 */
				/*new5 - see if our depth is so bad that even if diversity was 1 it still wouldn't help*/
				/*double value = prefs.alpha() * diversityMatch(subsetInstances, prefs, true);
				if( ( value + 1.0 - prefs.alpha() ) >= maxObjFnValue )
					value += ( (1.0-prefs.alpha() ) * depth(subsetInstances, prefs, true) );*/
				//old8 double value = ( (1.0-prefs.alpha() ) * depth(subsetInstances, prefs, true) );
				//new8 keep a total of depth without anything new and add on the depth for new instances
				double value = ( (1.0-prefs.alpha() )
								* depth(subsetInstances, prefs, true, baseSize, baseDepth, false) );
				if( ( value + prefs.alpha() ) >= maxObjFnValue )
					value += prefs.alpha() * diversityMatch(subsetInstances, prefs, true);
				
				subsetInstances.delete(subsetInstances.numInstances()-1); // removes from last index
				
				// if the objFnValue is greater, make this instance the new selected item
				if (value > maxObjFnValue) {
					maxObjFnValue = value;
					selectedItemsList.clear();
					selectedItemsList.add(candidateInstance);
				}
				// if the objFnValue is the same, add it to the selected items list
				else if (value == maxObjFnValue) {
					selectedItemsList.add(candidateInstance);
				}
			}
		}
		
		// if no item was found, throw an exception
		if (selectedItemsList.isEmpty()) {
			throw new IllegalStateException("No item selected.");
		}
	
		// add a random selected item to the subset
		subset.add(selectedItemsList.get(m_randGenerator.nextInt(selectedItemsList.size())));
		
		// return the new objective function value
		return maxObjFnValue;
	}

	
	/** Computes the DDPref objective function for a set of instances.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param prefs The preferences for the evaluation.
	 * @return The value of the objective function
	 */
	public static double computeObjFunction(Instances instances, Preferences prefs) {
		boolean normalizeByWeights = true;
		// return the value of the objective function
		return (prefs.alpha() * diversityMatch(instances, prefs, normalizeByWeights) 
			+ (1.0-prefs.alpha()) * depth(instances, prefs, normalizeByWeights));
	}
	
	/** Computes the diversity match for a set of instances to a set of preferences.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param prefs The preferences for the evaluation.
	 * @return The diversity.
	 */
	public static double diversityMatch(Instances instances, Preferences prefs) {
		return diversityMatch(instances, prefs, true);
	}
	
	/** Computes the diversity match for a set of instances to a set of preferences.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param prefs The preferences for the evaluation.
	 * @param normalizeByWeights <code>True</code> will normalize by the weight sum,
	 *                           <code>False</code> will normalize by the number of features.
	 * @return The diversity.
	 */
	private static double diversityMatch(Instances instances, Preferences prefs,
			boolean normalizeByWeights) {
		
		double beta = 1.0;
		
		double diversityError = 0;
		
		/*new5 store these vals*/
		int numAttribs = instances.numAttributes();
		int numInstances = instances.numInstances();
		int classIx = instances.classIndex();
		
		//new9 - get the double arrays of attribs all at the same time
		double[][] values = new double[numAttribs][numInstances];
		for( int idx = 0; idx < numInstances; idx++ ){
			double[] fVector = instances.instance(idx).toDoubleArray();
			for( int att = 0; att < numAttribs; att++ )
				values[att][idx] = fVector[att];
		}
		
		// loop through and compute the diversity
		for (int attIdx = 0; attIdx<numAttribs; attIdx++) {
			
			// skip the class attribute
			if (attIdx == classIx ) continue;
			
			// get the preference for this attribute
			Preference p = prefs.get(instances.attribute(attIdx).name());
			
			// skip if there isn't a preference
			if (p == null) {
				System.err.println("diversity() couldn't find preference for attribute \""+instances.attribute(attIdx).name()+"\".  Skipping this preference.");
				continue;
			}
			// skip preferences with a weight of 0
			if (p.weight() == 0) {
				//System.err.println("diversity() preference for attribute \""+instances.attribute(attIdx).name()+"\" has weight 0.  Skipping this preference.");
				continue;
			}
			// the error of the diversity is 1 minus the diversity fn evaluation
			double attributeDiversityError = 1.0-p.diversityFn(diversityFeature(instances, attIdx, beta, values[attIdx]));	
			// commpute the weighted sum of errors
			diversityError += p.weight() * attributeDiversityError;
		}
		
		// normalize the diversity by the preference weight sum
		if (normalizeByWeights) {
			diversityError /= prefs.weightSum();
		}
		// normalize by the number of features
		else {
			// if it has a class attribute
			if (classIx >= 0) {
				diversityError /= numAttribs - 1;
			}
			// if there isn't a class attribute
			else {
				diversityError /= numAttribs;
			}
		}
		
		// check that the diversity is in range
		if (!valueInRange(diversityError, 0.0, 1.0)) {
			throw new IllegalStateException("DiversityError must be between 0 and 1. Currently diversityError="+diversityError);
		}
		
		// return the diversity, which is 1 minus diversityError
		return 1.0-diversityError;
	}
	
	/** Computes the diversity of a particular feature.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param attIdx The attribute index of the feature.
	 * @param beta (CURRENTLY UNUSED)
	 * @return The diversity of the feature.
	 */
	private static double diversityFeature(Instances instances, int attIdx, double beta, double[] values) {
		// gather a list of the values for this particular attribute
		//old9 - now compute all attribs together
		//double[] values = instances.attributeToDoubleArray(attIdx);
		Arrays.sort(values);
		
		// compute the diversity of the feature
		double diversityFeature = 1.0 - skew(values);

		// check that the diversityMatch is in range
		if (!valueInRange(diversityFeature, 0.0, 1.0)) {
			throw new IllegalStateException("DiversityFeature must be between 0 and 1. Value is currently "+diversityFeature);
		}
		
		return diversityFeature;
	}
	
	/*
	 * This code incorporated the range function into the determination of the diversity
	 * 
	private double diversityFeature(ArrayList<Double> values, double maxRange, double beta) {
		return (beta * (1.0 - normalizedSkew(values)) 
				+ (1.0-beta) * rangeFn(range(values), maxRange));
	}
	
	private double range(ArrayList<Double> values) {
		Collections.sort(values);
		double min = values.get(0);
		double max = values.get(values.size()-1);
		return max - min;
	}
	
	private double rangeFn(double actualRange, double maxRange) {
		return 1.0;
	}
	*/
	
	/** Computes the normalized skew for a list of values.
	 * @param values The list of values.
	 * @return The normalized skew for the specified list.
	 *
	private double normalizedSkew(ArrayList<Double> values) {
		if (values.size() < 1) 
			throw new IllegalArgumentException("The list of values must include at least one element.");
		
		// sort the values and get the min and max
		Collections.sort(values);
		double min = values.get(0).doubleValue();
		double max = values.get(values.size()-1).doubleValue();
		
		// get the min and max skew values for this type of list
		double minSkew = minSkew(values.size(), min, max);
		double maxSkew = maxSkew(values.size(), min, max);
		
		// if the min and max skew are the same, the normalized skew is 1.
		if (minSkew == maxSkew) {
			return 1.0;
		}
		
		// compute and return the normalized skew.
		return (skew(values) - minSkew) / (maxSkew - minSkew);
	}
	*/
	
	/** Returns the minimum skew for an n-element list with the specified min and max.
	 * @param n The number of items in the list.
	 * @param min The min value in the list.
	 * @param max The max value in the list.
	 * @return The minimum skew for the specified list.
	 */
	private static double minSkew(int n, double min, double max) {
		return 0;
	}
	
	
	/** Returns the maximum skew for an n-element list with the specified min and max.
	 * The maximum skew occurs when all items (except the max item) are equal to min.
	 * @param n The number of items in the list.
	 * @param min The min value in the list.
	 * @param max The max value in the list.
	 * @return The maximum skew for the specified list.
	 */
	private static double maxSkew(int n, double min, double max) {
		double maxSkew = 0;
		double slope=(max-min) / (n-1);
		// assume all items except the last are equal to min
		for (int i=0; i < n-1; i++) {
			// maxSkew += Math.pow(expectedValue(i,min,max,n)-min,2);
			maxSkew += Math.pow(slope*i, 2);
		}
		return maxSkew;
	}
	
	
	
	/** Computes the skew for a sorted list of values.
	 * @param values The sorted list of values.
	 * @return The skew for the specified list.
	 */
	public static double skew(double[] values) {
		if (values.length < 1) 
			throw new IllegalArgumentException("The list of values must include at least one element.");
	
		// get the min and max
		double min = values[0];
		double max = values[values.length-1];
		
		// for completeness, the skew of a list composed of only one distinct value is 0,
		// as is a list with less than two values
		if (min == max || values.length <= 2) {
			return 0;
		} else {
			
		}
		
		// accumulate the skew going through each value in the list
		double skew = 0;
		double slope=(max-min) / (values.length-1);
		for (int i=0; i<values.length; i++) {
			double expectedValue = min + slope*i;
			skew += Math.pow(values[i] - expectedValue, 2);
		}
		
		// normalize the skew
		if (skew != 0) {
			skew /= maxSkew(values.length,min,max);
		}
			
		// check that the skew is in range
		if (!valueInRange(skew, 0.0, 1.0)) {
			throw new IllegalStateException("Skew must be between 0 and 1. Value is currently "+skew);
		}
		
		return skew;
	}
	
	
	/** Expected value of the ith item in an n-element sorted list with a linear distribution.
	 * @param i The index of the item in the list, zero-indexed.
	 * @param min The min value in the list.
	 * @param max The max value in the list.
	 * @param n The number of items in the list.
	 * @return The expected value of the ith item in this list with a linear distribution.
	 */
	private static double expectedValue(int i, double min, double max, double n) {
		return min + (max-min)*i/(n-1);
	}
	
	
	/** Computes the depth for a set of instances.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param prefs The preferences for the evaluation.
	 * @return The depth.
	 */
	public static double depth(Instances instances, Preferences prefs) {
		return depth(instances, prefs, true);
	}
	
	/**Used to store values for quality that have already been computed*/
	private static HashMap<String, HashMap<Double,Double> > qualities = new HashMap<String, HashMap<Double,Double> >();
	
	/** Computes the depth for a set of instances.
	 * 
	 * @param instances The set of instances to evaluate.
	 * @param prefs The preferences for the evaluation.
	 * @param normalizeByWeights <code>True</code> will normalize by the weight sum,
	 *                           <code>False</code> will normalize by the number of features.
	 * @return The depth.
	 */
	private static double depth(Instances instances, Preferences prefs, boolean normalizeByWeights){
		return depth(instances, prefs, normalizeByWeights, 0, 0, false );
	}
	
	//new8
	//Computes the depth given a partialSum for the first startIdx elements
	//Has a flag (which overrides normalizeByWieghts to return the unNormalized depth
	private static double depth(Instances instances, Preferences prefs, boolean normalizeByWeights,
								int startIdx, double partialSum, boolean unNormalized ) {
		double depth = partialSum; //new8
		
		// loop through each attribute
		/*new3 - store these values*/
		int numAttribs = instances.numAttributes();
		int numElems = instances.numInstances();
		int classIx  = instances.classIndex();
		for (int attIdx = 0; attIdx < numAttribs; attIdx++) {
			
			// skip the class attribute
			if (attIdx == classIx) continue;
			
			// get the preference for this attribute
			//new7 load the hashMap for this attribute
			String attName = instances.attribute(attIdx).name();
			HashMap<Double,Double> qualityMap = qualities.get(attName);
			if( qualityMap == null ){
				qualityMap = new HashMap<Double,Double>();
				qualities.put(attName,qualityMap);
			}
			Preference p = prefs.get(attName);
			
			// skip if there isn't a preference
			if (p == null) {
				System.err.println("depth() couldn't find preference for attribute \""+instances.attribute(attIdx).name()+"\".  Skipping this preference.");
				continue;
			}
			// skip preferences with a weight of 0
			if (p.weight() == 0) {
				//System.err.println("depth() preference for attribute \""+instances.attribute(attIdx).name()+"\" has weight 0.  Skipping this preference.");
				continue;
			}
			
			// loop through each instance
			for (int instIdx=startIdx; instIdx < numElems; instIdx++) {
				Instance inst = instances.instance(instIdx);
				// compute the contribution from the attribute of this object
				/*old6
				depth += p.weight() * quality(inst.value(attIdx), p);*/
				//new6 - store values we already computed to save future computations
				Double attVal = new Double( inst.value(attIdx) );
				Double qualityVal = (Double)qualityMap.get(attVal);
				if( qualityVal == null ){
					qualityVal = new Double( quality(attVal, p) );
					qualityMap.put(attVal,qualityVal);
				}
			//double qualityVal = quality(inst.value(attIdx), p);
				depth += p.weight()*qualityVal;
			}

		}
		
		//new8 return unNormalized depth if requested
		if( unNormalized)
			return depth;
		
		// normalize the depth by the preference weight sum
		if (normalizeByWeights) {
			depth /= prefs.weightSum();
		}
		// normalize by the number of features
		else {
			// if it has a class attribute
			if (classIx >= 0) {
				depth /= numAttribs - 1;
			}
			// if there isn't a class attribute
			else {
				depth /= numAttribs;
			}
		}
		
		// normalize by the number of objects
		depth /= numElems;
		
		// check that the depth is in range
		/*new4 inline of if (!valueInRange(depth, 0.0, 1.0))*/
		if ( !( depth >= -0.0001 && depth <= 1.0001 ) ) {
			throw new IllegalStateException("Depth must be between 0 and 1. Currently depth="+depth);
		}
		
		// return the depth
		return depth;
	}

	/** Computes the quality of a particular value according to a preference.
	 * 
	 * @param value The value to evaluate.
	 * @param pref The preference for the evaluation.
	 * @return The quality of the value.
	 */
	public static double quality(double value, Preference pref) {
		// use the quality function if one exists
		if (pref.hasQualityFn()) {
			return pref.qualityFn(value);
		}
		// otherwise, the quality is 1
		else {
			return 1.0;
		}
	}
	
	
		
	/** Checks that a value is within the specified range.
	 * @param value The value to check.
	 * @param minValue The minimum value allowed.
	 * @param maxValue The maximum value allowed.
	 * @return <code>true</code> if minValue <= value <= maxValue, <code>false</code> otherwise.
	 */
	private static boolean valueInRange(double value, double minValue, double maxValue) {
		double EPSILON = 1.0E-4;
		return (value >= (minValue - EPSILON) && value <= (maxValue + EPSILON));
	}
	

	public static enum SelectionMethod {wrapperSelectSubset, greedySelectSubset, exhaustiveSelectSubset, randomSelectSubset, topkSelectSubset};
	
	
	/** Generates the cross measure for a set of preferences.
	 */
	public String crossMeasure(int k, ArrayList<Preferences> prefsList, SelectionMethod selectionMethod) {
		String str = "";
		// generate subsets using each of the preferences
		ArrayList<Instances> selectedSubsets = new ArrayList<Instances>(prefsList.size()+1);
		for (Preferences prefs : prefsList) {
			Instances selectedInstances = null;
			switch(selectionMethod) {
			case wrapperSelectSubset:
				selectedInstances = wrapperSelectSubset(k, prefs); break;
			case greedySelectSubset:
				selectedInstances = greedySelectSubset(k, prefs); break;
			case exhaustiveSelectSubset:
				selectedInstances = exhaustiveSelectSubset(k, prefs); break;
			case topkSelectSubset:
				selectedInstances = topSelectSubset(k,prefs); break;
			case randomSelectSubset:
				selectedInstances = randomSelectSubset(k, prefs); break;
			default:
				throw new IllegalStateException("Unhandled selection method.");	
			}
			selectedSubsets.add(selectedInstances);
		}
		selectedSubsets.add(randomSelectSubset(k,null,null));
		
		// for each generated subset, including random, evaluate the subset against each preference
		// output header
		str += "\t";
		for (Preferences prefs : prefsList) {
			str += "\t"+prefs.name();
		}
		str += "\tRandom\n";
		// output data table
		int index = 0;
		for (Instances subset : selectedSubsets) {
			// output preference that generated the subset
			if (index < prefsList.size())
				str += prefsList.get(index).name();
			else 
				str += "Random";
			for (Preferences prefs : prefsList) {
				str += "\t"+computeObjFunction(subset,prefs);
			}
			index++;
			str += "\n";
		}
		return str;
	}
	
	
	
	private static double[] parseDoubleArray(String strDoubles) {
		String[] strDoublesArray = strDoubles.split(",");
		double[] doubles = new double[strDoublesArray.length];
		for (int i=0; i<doubles.length; i++) {
			doubles[i] = Double.parseDouble(strDoublesArray[i]);
		}
		return doubles;
	}
	
	private static int[] parseIntegerArray(String strInts) {
		String[] strIntsArray = strInts.split(",");
		int[] ints = new int[strIntsArray.length];
		for (int i=0; i<ints.length; i++) {
			ints[i] = Integer.parseInt(strIntsArray[i]);
		}
		return ints;
	}
	
	private static String printUsage() {
		String s = "Usage: DDPref <options>\n";
		s += "-D <dataset filename> (required if -G unspecified)\n";
		s += "-G <dataset generator> (required if -D unspecified)\n";
		s += "-n <number of instances to use from the data set (required if -G specified)\n";
		s += "-P <preference filename> (one required, multiple possible)\n";
		s += "-k <list of values for the size of the subset> (e.g. -k 5,8,10) (required)\n";
		s += "-a <list of values for alpha> (e.g. -a 0.5,1) (default: 0.5)\n";
		s += "-t <number of trials> (default: 1)\n";
		s += "-r <random seed>";
		return s;
	}
	
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		// command line argument variables
		BlocksGenerator dataGenerator = null;
		int numDataItems = -1;
		int[] kList = null;
		double[] alphaList = new double[]{0.5};
		Class dataGeneratorClass = null;
		int numTrials = 1;
		Instances data = null;
		ArrayList<Preferences> prefsList = new ArrayList<Preferences>();
		long randomSeed = System.currentTimeMillis();
		
		boolean verbose = false;
		
		// parse the command line arguments
		try {
			String optionString = Utils.getOption('G',args);
			if (optionString.length() != 0) {
				dataGeneratorClass = ddpref.data.BlocksGenerator.class;//arguments[0];
			}
			
			optionString = Utils.getOption('n',args);
			if (optionString.length() != 0) {
				numDataItems = Integer.parseInt(optionString);
			}
			
			kList = parseIntegerArray(Utils.getOption('k',args));
			
			optionString = Utils.getOption('a',args);
			if (optionString.length() != 0) {
				alphaList = parseDoubleArray(optionString);
			}
			
			optionString = Utils.getOption('t',args);
			if (optionString.length() != 0) {
				numTrials = Integer.parseInt(optionString);
			}
			
			optionString = Utils.getOption('r',args);
			if (optionString.length() != 0) {
				randomSeed = Integer.parseInt(optionString);
			}
			
			optionString = Utils.getOption('D',args);
			if (optionString.length() != 0) {
				data = new Instances(new FileReader(new File(optionString)));
			}
			
			optionString = Utils.getOption('P',args);
			while (optionString.length() != 0) {
				prefsList.add(new Preferences(optionString,new File(optionString)));
				optionString = Utils.getOption('P',args);
			}
		} catch(Exception e) {
			System.err.println(e.getMessage());
			System.err.println(printUsage());
			System.exit(1);
		}
		
		// error check command line arguments
		// TODO Check these!!
		try {
			if (kList == null) throw new Exception("k unspecified");
			for (double k : kList)
				if (!valueInRange(k,1,Integer.MAX_VALUE)) throw new Exception("k value out of range");
			for (double a : alphaList)
				if (!valueInRange(a,0,1)) throw new Exception("alpha out of range");
			if (!valueInRange(numTrials,1,Integer.MAX_VALUE)) throw new Exception("numTrials out of range");
			
			if (data == null && dataGeneratorClass == null) throw new Exception("-D or -G must be specified");
			if (data != null && dataGeneratorClass != null) throw new Exception("Only specify one of -D or -G");
			if (data == null) {
				dataGenerator = new BlocksGenerator(randomSeed);
			}
			if (prefsList.size() < 1) throw new Exception ("At least one -P must be specified.");
			if (numDataItems < 0 && dataGeneratorClass != null) throw new Exception("-n must be specified when -G is specified");
			if (numDataItems < 0 && data != null) numDataItems = data.numInstances(); 
		} catch(Exception e) {
			System.err.println("Required value unspecified or out of range: "+e.getMessage());
			System.err.println(printUsage());
			System.exit(1);
		}
		
		try {
			
			Instances selectedSubset = null;
			Instances experimentData = null;
			
			String crossMeasure = "";
			
			// Run the experiment over all values of alpha, and k.
			for (double alpha : alphaList) {
				for (int k : kList) {
					
					double[][] cmSum_Greedy = null;
					double[][] cmSum_Wrapper = null;
					double[][] cmSum_Exhaustive = null;
					double[][] cmSum_TopK = null;
					double[][] cmSum_Random = null;
					
					boolean trackCrossMeasure = false; //was true - false to speed it up
					
					// for a given alpha, and k, generate the subset for each preference set
					for (Preferences prefs : prefsList) {
						
						double objFnSum_Greedy = 0;
						double objFnSum_Wrapper = 0;
						double objFnSum_Random = 0;
						double objFnSum_TopK = 0;
						double objFnSum_Exhaustive = 0;
						double timeSum_Greedy = 0;
						double timeSum_Wrapper = 0;
						double timeSum_Random = 0;
						double timeSum_TopK = 0;
						double timeSum_Exhaustive = 0;
						
						Random randGenerator = new Random(randomSeed+1); // reset the random generator
						
						for (int trial = 0; trial < numTrials; trial++) {
							
							// generate the data if there aren't specified examples
							if (dataGenerator != null) {
								experimentData = dataGenerator.generateInstances(numDataItems);
							}
							// otherwise shuffle and trim the data set
							else {
								data.randomize(randGenerator);
								experimentData = new Instances(data);
								for (int i=0; i<numDataItems; i++) {
									experimentData.add(data.instance(i));
								}
							}
							
							// initialize the preference engine
							DDPref ddpref = new DDPref();
							ddpref.setData(experimentData);
							prefs.setAlpha(alpha);
							ddpref.setPreferences(prefs);
							if (verbose) System.out.println();
							if (verbose) System.out.println("Running experiments: trial="+(trial+1)+", alpha=" + alpha + ", k=" + k+", Prefs="+prefs.name());
							
							/*
							 * // set up result file output String resultsFilename =
							 * outputDirectory+"alpha"+alpha+"n"+n+"k"+k+"pref"+pref+"alg"+alg+".dat";
							 * BufferedWriter resultsWriter = new BufferedWriter(new
							 * FileWriter(resultsFilename)); resultsWriter.write(";;
							 * alpha="+alpha+" n="+n+" k="+k+" pref="+p+"
							 * alg="+alg+"\n"); resultsWriter.write(";; Data format:
							 * Obj-fn-value CPU-time\n"); // set up blocks file
							 * output String blocksFilename =
							 * outputDirectory+"BLalpha"+alpha+"n"+n+"k"+k+"pref"+pref+"alg"+alg+".dat";
							 * BufferedWriter blocksWriter = new BufferedWriter(new
							 * FileWriter(blocksFilename)); blocksWriter.write(";;
							 * alpha="+alpha+" n="+n+" k="+k+" pref="+p+"
							 * alg="+alg+"\n"); blocksWriter.write(";; Data format:
							 * (block1 block2 ...)\n");
							 */
							
							double time = System.currentTimeMillis();
							selectedSubset = ddpref.randomSelectSubset(k);
							time = System.currentTimeMillis() - time;
							double objFnValue = ddpref.getObjFnValue();
							objFnSum_Random += objFnValue;
							timeSum_Random += time;
							if (verbose) System.out.println("Random Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							
							time = System.currentTimeMillis();
							selectedSubset = ddpref.topSelectSubset(k);
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_TopK += objFnValue;
							timeSum_TopK += time;
							if (verbose) System.out.println("TopK Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							
							time = System.currentTimeMillis();
							selectedSubset = ddpref.greedySelectSubset(k);
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_Greedy += objFnValue;
							timeSum_Greedy += time;
							if (verbose) System.out.println("Greedy Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							//System.out.println("");
							
							
							time = System.currentTimeMillis();
							selectedSubset = ddpref.wrapperSelectSubset(k);           //This takes 99% of time
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_Wrapper += objFnValue;
							timeSum_Wrapper += time;
							if (verbose) System.out.println("Wrapper Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							//System.out.println("");
							
							
							//temporarily excluded to make experiment managable
							//commented stuff maked with //exhaustive
							//time = System.currentTimeMillis();
							//selectedSubset = ddpref.exhaustiveSelectSubset(k);
							//time = System.currentTimeMillis() - time;
							//objFnValue = ddpref.getObjFnValue();
							//objFnSum_Exhaustive += objFnValue;
							//timeSum_Exhaustive += time;
							//if (verbose) System.out.println("Exhaustive Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							
							
							if (verbose) System.out.println();
							/*
							 * if (resultsWriter != null) {
							 * resultsWriter.write(objFnValue+" "+time+"\n"); } if
							 * (blocksWriter != null) {
							 * blocksWriter.write(subset+"\n"); }
							 * timeAndRecord(resultsWriter, blockWriter, alg, data,
							 * features, pref, k, alpha);
							 *  // close the file writers resultsWriter.close();
							 * blocksWriter.close();
							 */
							
							// generate the cross-product of the algorithms on the preferences
							if (trackCrossMeasure) {

								if (verbose) System.out.println("Cross-Measure for Random");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.randomSelectSubset);
								cmSum_Random = updateCMSum(crossMeasure,cmSum_Random);
								if (verbose) System.out.println(crossMeasure);

								if (verbose) System.out.println("Cross-Measure for TopK");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.topkSelectSubset);
								cmSum_TopK = updateCMSum(crossMeasure,cmSum_TopK);
								if (verbose) System.out.println(crossMeasure);
								
								if (verbose) System.out.println("Cross-Measure for Greedy");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.greedySelectSubset);
								cmSum_Greedy = updateCMSum(crossMeasure,cmSum_Greedy);
								if (verbose) System.out.println(crossMeasure);
								
								if (verbose) System.out.println("Cross-Measure for Wrapper");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.wrapperSelectSubset);
								cmSum_Wrapper = updateCMSum(crossMeasure,cmSum_Wrapper);
								if (verbose) System.out.println(crossMeasure);
								
								//exhaustive
								/*if (verbose) System.out.println("Cross-Measure for Exhaustive");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.exhaustiveSelectSubset);
								cmSum_Exhaustive = updateCMSum(crossMeasure,cmSum_Exhaustive);
								if (verbose) System.out.println(crossMeasure);*/
								if (verbose) System.out.println();
							}
						}
						System.out.println("For Alpha = "+alpha+" k = "+k+" prefs = "+prefs.name()+" over "+numTrials+" trials:");
						System.out.println("Summary info for Random:     objFn = "+(objFnSum_Random/numTrials)+"  time = "+((timeSum_Random/numTrials)/1000));
						System.out.println("Summary info for TopK:       objFn = "+(objFnSum_TopK/numTrials)+"  time = "+((timeSum_TopK/numTrials)/1000));
						System.out.println("Summary info for Greedy:     objFn = "+(objFnSum_Greedy/numTrials)+"  time = "+((timeSum_Greedy/numTrials)/1000));
						System.out.println("Summary info for Wrapper:    objFn = "+(objFnSum_Wrapper/numTrials)+"  time = "+((timeSum_Wrapper/numTrials)/1000));
						//exhaustive
						//System.out.println("Summary info for Exhaustive: objFn = "+(objFnSum_Exhaustive/numTrials)+"  time = "+((timeSum_Exhaustive/numTrials)/1000));
						
						System.out.println();
						trackCrossMeasure = false; // stop tracking cross measure after one run of trials
					}
					System.out.println("For Alpha = "+alpha+" k = "+k+" over "+numTrials+" trials:");
					System.out.println("Summary info for Cross-Measure Random over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Random,numTrials));
					System.out.println("Summary info for Cross-Measure TopK over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_TopK,numTrials));
					System.out.println("Summary info for Cross-Measure Greedy over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Greedy,numTrials));
					System.out.println("Summary info for Cross-Measure Wrapper over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Wrapper,numTrials));
					//exhaustive
					//System.out.println("Summary info for Cross-Measure Exhaustive over "+numTrials+" trials.");
					//System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Exhaustive,numTrials));
				}
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
		
	}
	
	private static String generateCMMeanOutput(String cmString, double[][] cmSum, int numTrials) {
		String[] cmStringLines = cmString.split("\n");
		String str = cmStringLines[0]+"\n";
		for (int i=1; i<cmStringLines.length; i++) {
			String[] lineSplit = cmStringLines[i].split("\t");
			str += lineSplit[0];
			for (int j=1; j<lineSplit.length; j++) {
				str += "\t" + (cmSum[i-1][j-1] / numTrials);
			}
			str += "\n";
		}
		return str;
	}
	
	private static double[][] updateCMSum(String cmString,double[][] cmSum) {
		String[] cmStringLines = cmString.split("\n");
		for (int i=1; i<cmStringLines.length; i++) {
			String[] lineSplit = cmStringLines[i].split("\t");
			if (cmSum == null) cmSum = new double[cmStringLines.length][lineSplit.length];
			for (int j=1; j<lineSplit.length; j++) {
				cmSum[i-1][j-1] += Double.parseDouble(lineSplit[j]);
			}
		}
		return cmSum;
	}

}
