/* DDPrefLearner.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;


import weka.classifiers.Classifier;
import weka.core.*;
import weka.estimators.*;

import java.io.*;
import java.lang.Character.Subset;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Random;
import java.util.Scanner;
import java.util.TreeSet;

import org.apache.bsf.util.event.adapters.java_awt_event_ActionAdapter;
import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;

import ddpref.data.BlocksGenerator;

public class DDPrefLearner {
	
	static final double ATTRIBUTE_PRECISION = 0.01;   //0.1
 	static final double DIVERSITY_PRECISION = 0.01;  //0.0001
	
	private double[] m_weights = null;
	private double m_alpha = -1;
	private double m_minFunctionValue = -1;
	private Preferences m_preferences = null;
	

	private boolean m_learnDiversityFunction = false;
	public void setUseKDEDiversityLearning(boolean useKDEDiversityEstimation){
		m_learnDiversityFunction = useKDEDiversityEstimation;
	}
	
	private boolean m_learningAlpha = true;
	public void setLearningAlpha(boolean value) {
		m_learningAlpha = value;
	}
	
	public void setAlpha(double alpha) {
		m_alpha = alpha;
	}
	
	
	private boolean m_learningQualityFns = true;
	private boolean learningQualityFns() { return m_learningQualityFns; }
	public void setLearningQualityFn(boolean value) {
		m_learningQualityFns = value;
	}
	private Preferences m_qualityFnPrefs = null;
	public void setQualityFns(Preferences prefs) {
		m_qualityFnPrefs = prefs;
	}
	
	/** learn the quality functions for each attribute value using kernel density estimators
	 * @return an array of quality functions
	 */
	private KernelDensityEstimator[] learnQualityFunctions(Collection<Instances> trainingSubsets) {
		KernelDensityEstimator[] qualityFunctions = null;
		for (Instances subset : trainingSubsets) {
			// initialize the quality functions if they haven't been already
			if (qualityFunctions == null) {
				qualityFunctions = new KernelDensityEstimator[subset.numAttributes()];
				int numAttributes = subset.numAttributes();
				for (int i=0; i<numAttributes; i++) {
					//qualityFunctions[i] = new QualityFnKernelEstimator(ATTRIBUTE_PRECISION);
					//qualityFunctions[i] = new SingleStepEstimator();
					qualityFunctions[i] = new QualityFnKernelDensityEstimator(ATTRIBUTE_PRECISION);
					
				}
			}
			// accumulate the values from each attribute over all instances
			int subsetNumInstances = subset.numInstances();
			int subsetNumAttributes = subset.numAttributes();
			for (int instIdx=0; instIdx < subsetNumInstances; instIdx++) {
				Instance inst = subset.instance(instIdx);
				for (int i=0; i<subsetNumAttributes; i++) {
					qualityFunctions[i].addValue(inst.value(i), 1.0);
				}
			}
		}
		return qualityFunctions;
	}
	
	
	/** Learn the diversities from the training subsets
	 * @return the expected diversities
	 */
	private double[] learnDiversities(Collection<Instances> trainingSubsets) {
		
		double[] diversities = null;
		
		// compute the diversity for each attribute
		for (Instances subset : trainingSubsets) {
			// initialize the diversities if they haven't been already
			if (diversities == null) {
				diversities = new double[subset.numAttributes()];
			}
			
			// loop over each attribute
			int subsetNumAttributes = subset.numAttributes();
			for (int attIdx=0; attIdx<subsetNumAttributes; attIdx++) {
				double[] attValues = subset.attributeToDoubleArray(attIdx);
				// accumulate the expected diversity
				diversities[attIdx] += diversityFeature(attValues);
			}
		}
		
		// compute the expected diversity
		for (int attIdx=0; attIdx<diversities.length; attIdx++) {
			diversities[attIdx] /= (double) trainingSubsets.size();
		}
		
		return diversities;
	}
	
	/** Learn diversity functions from the training subsets
	 * @return the expected diversities
	 */
	private KernelDensityEstimator[] learnDiversityFunctions(Collection<Instances> trainingSubsets) {
		
		KernelDensityEstimator[] diversityFunctions = null;
		for (Instances subset : trainingSubsets) {
			// initialize the diversity functions if they haven't been already
			int subsetNumAttributes = subset.numAttributes();
			if (diversityFunctions == null) {
				diversityFunctions = new KernelDensityEstimator[subset.numAttributes()];
				for (int i=0; i<subsetNumAttributes; i++) {
					//qualityFunctions[i] = new QualityFnKernelEstimator(ATTRIBUTE_PRECISION);
					//qualityFunctions[i] = new SingleStepEstimator();
					diversityFunctions[i] = new QualityFnKernelDensityEstimator(DIVERSITY_PRECISION);
					
				}
			}
			
			// loop over each attribute
			for (int attIdx=0; attIdx<subsetNumAttributes; attIdx++) {
				double[] attValues = subset.attributeToDoubleArray(attIdx);
				// accumulate the expected diversity
				diversityFunctions[attIdx].addValue(diversityFeature(attValues),1.0);
			}
		}

		return diversityFunctions;
	}
	
	
	static public double diversityFeature(double[] values) {
		// return the diversity
		double[] tempValues = new double[values.length];
		System.arraycopy(values,0,tempValues,0,values.length);
		Arrays.sort(tempValues);
		//return (DDPref.skew(values));
		return (1.0 - DDPref.skew(tempValues));
	}
	
	
	/** Trains the learner based on a set of subsets.
	 * Assumes all training subsets have a target objective function value of 1.0.
	 * @param trainingSubsets The set of subsets to be used in training
	 */
	public void train(Collection<Instances> trainingSubsets) {
		double[] objFnValues = new double[trainingSubsets.size()];
		for (int i=0; i<objFnValues.length; i++) {
			objFnValues[i] = 1.0;
		}
		train(trainingSubsets, objFnValues);
	}
	
	/** Trains the learner based on a set of subsets.
	 * @param trainingSubsets The set of subsets to be used in training.
	 * @param 
	 */
	public void train(Collection<Instances> trainingSubsets, double[] objFnValues) {
		
		// learn the quality function for each attribute value using kernel density estimators
		KernelDensityEstimator[] qualityFunctions = null;
		if (learningQualityFns()) 
			qualityFunctions = learnQualityFunctions(trainingSubsets);
		
		// learn the desired diversity
		double[] diversities = null;
		KernelDensityEstimator[] diversityFunctions = null;
		if (m_learnDiversityFunction) {
			diversityFunctions = learnDiversityFunctions(trainingSubsets);
		} else {
			diversities = learnDiversities(trainingSubsets);
		}
		
		// compute the mean quality and diversity match for each attribute of each subset
		double[][] meanQualities = new double[trainingSubsets.size()][];
		double[][] diversityMatches = new double[trainingSubsets.size()][];
		int subsetIdx=-1;
		int numAttributes = -1;
		// loop over each subset
		for (Instances subset : trainingSubsets) {
			subsetIdx++;
			
			// compute the mean quality for each attribute of each subset
			// initialize the summedQualities to zero
			numAttributes = subset.numAttributes();
			meanQualities[subsetIdx] = new double[numAttributes];
			for (int attIdx=0; attIdx<numAttributes; attIdx++) {
				meanQualities[subsetIdx][attIdx] = 0;
			}
			// accumulate the qualities for each attribute over the subset's members
			int numInstances = subset.numInstances();
			for (int instIdx=0; instIdx<numInstances; instIdx++) {
				Instance inst = subset.instance(instIdx);
				for (int attIdx=0; attIdx<numAttributes; attIdx++) {
					if (learningQualityFns()) // use the learned quality functions
						meanQualities[subsetIdx][attIdx] += qualityFunctions[attIdx].getProbability(inst.value(attIdx));
					else // use the quality functions given in the preferences
						meanQualities[subsetIdx][attIdx] += DDPref.quality(inst.value(attIdx), m_qualityFnPrefs.get(inst.attribute(attIdx).name()));
				}
			}
			// divide the qualities by the number of items
			for (int attIdx=0; attIdx<numAttributes; attIdx++) {
				meanQualities[subsetIdx][attIdx] /= (double)subset.numInstances();
			}
			
			// compute the squared diversity error for each attribute of each subset
			diversityMatches[subsetIdx] = new double[numAttributes];
			// loop over each attribute
			for (int attIdx=0; attIdx<numAttributes; attIdx++) {
				// compute the diversity of this subset along this attribute
				double diversity = diversityFeature(subset.attributeToDoubleArray(attIdx));
				double diversityError = Double.NaN;
				// compute and store the squared diversity error
				if (m_learnDiversityFunction) {
					// the error of the diversity is 1 minus the diversity fn evaluation
					diversityError = 1.0-diversityFunctions[attIdx].getProbability(diversity);	
				} else {
					diversityError = diversities[attIdx] - diversity;
				}
				diversityMatches[subsetIdx][attIdx] = 1.0 - Math.pow(diversityError, 2);
			}
		}
		
		// prepare to optimize the objective function
		DDPrefOptimization opt = (m_learningAlpha) ? 
				new DDPrefOptimization(meanQualities, diversityMatches, objFnValues) :
				new DDPrefOptimization(meanQualities, diversityMatches, objFnValues, m_alpha);
		
		// set initial weights
		double[] weights = (m_learningAlpha) ? new double[numAttributes+1] : new double[numAttributes];
		for (int i=0; i<(weights.length-1); i++) {
			weights[i] = 1.0; // initial weight of 1 for every attribute
		}
		if (m_learningAlpha) {
			weights[weights.length-1] = 0.5; // initial alpha of 0.5
		}
			
		// set the lower and upper bounds to [0,1]
		// 1st row is lower bounds, 2nd is upper
		double[][] constraints = new double[2][weights.length];
		for (int i=0; i<weights.length; i++) {
			constraints[0][i] = 1.0E-10;
			constraints[1][i] = 1.0;
		}
/*
 * ADDED BY JC MONTMINY
 * if true, learn the weights.
 * if false, bypass the weight learning altogether.
 * END COMMENTS ADDED BY JC MONTMINY
 * 
 */
if (true){
		// minimize the function, 200 iterations as default
		try {
			weights = opt.findArgmin(weights, constraints);
			int iterationCount = 1;
			while(weights == null){  // 200 iterations are not enough
				System.out.println("At "+ (200 * iterationCount)+ " iterations.");
				iterationCount++;
				weights = opt.getVarbValues();  // try another 200 iterations
				weights = opt.findArgmin(weights, constraints);
			}
		} catch(Exception e) {
			System.err.println("Training is incomplete.  Function optimization failed.");
			System.err.println("Message: "+e.getMessage());
		}
}
		// the minimal function value
		m_minFunctionValue = opt.getMinFunction();

		// store the final weights and alpha
		m_weights = new double[numAttributes];
		System.arraycopy(weights,0,m_weights,0,numAttributes);
		if (m_learningAlpha) {
			m_alpha = weights[weights.length-1];
		}
		
		// normalize the weights
		double maxWeight = Double.NEGATIVE_INFINITY;
		for (double weight : m_weights) {
			maxWeight = (weight>maxWeight)?weight:maxWeight;
		}
		for (int i=0; i<m_weights.length; i++) {
			m_weights[i] /= maxWeight;
		}
		
		//System.out.println("Final Minimized Fn Value = "+m_minFunctionValue+" ["+arrayToString(m_weights)+"]");
		

		// construct the preferences
		// extract the attribute information from the data
		Instances repData = trainingSubsets.iterator().next();
		Attribute[] attributes = new Attribute[numAttributes];
		for (int i=0; i<attributes.length; i++) {
			attributes[i] = repData.attribute(i);
		} 
		if (learningQualityFns()) // use the learned quality functions
			constructPreferences(attributes, m_alpha, m_weights, diversities, diversityFunctions, qualityFunctions);
		else // use the quality functions given in the preferences
			constructPreferences(attributes, m_alpha, m_weights, diversities, diversityFunctions, m_qualityFnPrefs);

		
		//System.out.println(m_preferences);
	}
	
	private void constructPreferences(Attribute[] attributes,
			double alpha, double[] attWeights, double[] attDiversities, KernelDensityEstimator[] attDiversityFns, Estimator[] attQualityEstimators) {
		
		m_preferences = new Preferences("Learned by DDPrefLearner");
		m_preferences.setAlpha(alpha);
		// loop through each attribute
		for (int attIdx=0; attIdx<attributes.length; attIdx++) {
			Preference pref = new Preference();
			pref.setAttributeName(attributes[attIdx].name());
			pref.setWeight(attWeights[attIdx]);
			if (m_learnDiversityFunction)
				pref.setDiversityFn(attDiversityFns[attIdx]);
			else
				pref.setDiversityFn(attDiversities[attIdx]);
			pref.setQualityFn(attQualityEstimators[attIdx]);
			m_preferences.add(pref);
		}
	}
	
	private void constructPreferences(Attribute[] attributes,
			double alpha, double[] attWeights, double[] attDiversities, KernelDensityEstimator[] attDiversityFns, Preferences qualityFnPrefs) {
		
		m_preferences = new Preferences("Learned by DDPrefLearner");
		m_preferences.setAlpha(alpha);
		// loop through each attribute
		for (int attIdx=0; attIdx<attributes.length; attIdx++) {
			Preference qualityFnPref = qualityFnPrefs.get(attributes[attIdx].name());
			 // clone the preference with the quality function, then change key parts
			Preference pref = new Preference(qualityFnPref);
			pref.setAttributeName(attributes[attIdx].name());
			pref.setWeight(attWeights[attIdx]);
			if (m_learnDiversityFunction)
				pref.setDiversityFn(attDiversityFns[attIdx]);
			else
				pref.setDiversityFn(attDiversities[attIdx]);
			m_preferences.add(pref);
		}
	}
	
	
	public Preferences getPreferences() {
		return m_preferences;
	}
	
	
	
	private static String printUsage() {
		String s = "Usage: DDPref <options>\n";
		s += "-D <dataset filename> (required if -G unspecified)\n";
		s += "-G <dataset generator> (required if -D unspecified)";
		s += "-n <number of instances to use from the data set (required if -G specified)";
		s += "-P <preference filename> (one required, multiple possible)\n";
		s += "-k <list of values for the size of the subset> (e.g. -k 5,8,10) (required)\n";
		s += "-a <list of values for alpha> (e.g. -a 0.5,1) (default: 1)\n";
		s += "-t <number of trials> (default: 1)";
		s += "-r <random seed>";
		return s;
	}
	
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		// command line argument variables
		BlocksGenerator dataGenerator = null;
		int numDataItems = -1;
		int[] kList = null;
		double[] alphaList = new double[]{1};
		Class dataGeneratorClass = null;
		int numTrials = 1;
		Instances data = null;
		ArrayList<Preferences> prefsList = new ArrayList<Preferences>();
		long randomSeed = System.currentTimeMillis();
		
		boolean verbose = false;
		
		// parse the command line arguments
		try {
			String optionString = Utils.getOption('G',args);
			if (optionString.length() != 0) {
				dataGeneratorClass = ddpref.data.BlocksGenerator.class;//arguments[0];
			}
			
			optionString = Utils.getOption('n',args);
			if (optionString.length() != 0) {
				numDataItems = Integer.parseInt(optionString);
			}
			
			kList = parseIntegerArray(Utils.getOption('k',args));
			
			optionString = Utils.getOption('a',args);
			if (optionString.length() != 0) {
				alphaList = parseDoubleArray(optionString);
			}
			
			optionString = Utils.getOption('t',args);
			if (optionString.length() != 0) {
				numTrials = Integer.parseInt(optionString);
			}
			
			optionString = Utils.getOption('r',args);
			if (optionString.length() != 0) {
				randomSeed = Integer.parseInt(optionString);
			}
			
			optionString = Utils.getOption('D',args);
			if (optionString.length() != 0) {
				data = new Instances(new FileReader(new File(optionString)));
			}
			
			optionString = Utils.getOption('P',args);
			while (optionString.length() != 0) {
				prefsList.add(new Preferences(optionString,new File(optionString)));
				optionString = Utils.getOption('P',args);
			}
		} catch(Exception e) {
			System.err.println(e.getMessage());
			System.err.println(printUsage());
			System.exit(1);
		}
		
		// error check command line arguments
		// TODO Check these!!
		try {
			if (kList == null) throw new Exception("k unspecified");
			for (double k : kList)
				if (!valueInRange(k,1,Integer.MAX_VALUE)) throw new Exception("k value out of range");
			for (double a : alphaList)
				if (!valueInRange(a,0,1)) throw new Exception("alpha out of range");
			if (!valueInRange(numTrials,1,Integer.MAX_VALUE)) throw new Exception("numTrials out of range");
			
			if (data == null && dataGeneratorClass == null) throw new Exception("-D or -G must be specified");
			if (data != null && dataGeneratorClass != null) throw new Exception("Only specify one of -D or -G");
			if (data == null) {
				dataGenerator = new BlocksGenerator(randomSeed);
			}
			if (prefsList.size() < 1) throw new Exception ("At least one -P must be specified.");
			if (numDataItems < 0 && dataGeneratorClass != null) throw new Exception("-n must be specified when -G is specified");
			if (numDataItems < 0 && data != null) numDataItems = data.numInstances(); 
		} catch(Exception e) {
			System.err.println("Required value unspecified or out of range: "+e.getMessage());
			System.err.println(printUsage());
			System.exit(1);
		}
		
		try {
			
			Instances experimentData = null;
			
			String crossMeasure = "";
			
			// Run the experiment over all values of alpha, and k.
			for (double alpha : alphaList) {
				for (int k : kList) {

					
					// for a given alpha, and k, generate the subset for each preference set
					for (Preferences prefs : prefsList) {
						
						Random randGenerator = new Random(randomSeed+1); // reset the random generator
						
						SummaryStatistics objFn_Greedy = new SummaryStatisticsImpl();
						SummaryStatistics objFn_Wrapper = new SummaryStatisticsImpl();
						SummaryStatistics objFn_Random = new SummaryStatisticsImpl();
						SummaryStatistics objFn_TopK = new SummaryStatisticsImpl();
						SummaryStatistics objFn_Learned = new SummaryStatisticsImpl();
						
						for (int trial = 0; trial < numTrials; trial++) {
							
							// generate the data if there aren't specified examples
							if (dataGenerator != null) {
								//experimentData = dataGenerator.generateInstances(numDataItems);
							}
							// otherwise, can't use this learning setup
							else {
								System.err.println("Must use data generation, currently.");
								System.exit(1);
							}
							
							// initialize the preference engine
							DDPref ddpref = new DDPref();
							prefs.setAlpha(alpha);
							ddpref.setPreferences(prefs);

							if (verbose) System.out.println();
							if (verbose) System.out.println("Running experiments: trial="+(trial+1)+", alpha=" + alpha + ", k=" + k+", Prefs="+prefs.name());
							
							
							// get a set of training data from ddpref
							int numTrainingSubsets = 20;
							ArrayList<Instances> trainingSubsets = new ArrayList<Instances>(numTrainingSubsets);
							for (int i=0; i<numTrainingSubsets; i++) {
								Instances tempData = dataGenerator.generateInstances(numDataItems);
								ddpref.setData(tempData);
								trainingSubsets.add(ddpref.wrapperSelectSubset(k));
								//System.out.println("Training Fn Value = "+ddpref.getObjFnValue());
							}
							
							// get evaluation data from ddpref
							Instances evaluationData = dataGenerator.generateInstances(numDataItems);
							
							// use the training data to learn the preference function
							DDPrefLearner ddprefLearner = new DDPrefLearner();
							ddprefLearner.setAlpha(alpha);
							ddprefLearner.setLearningAlpha(false);
							ddprefLearner.train(trainingSubsets);
							Preferences learnedPrefs = ddprefLearner.getPreferences();
							
							// evaluate the learned preferences against the original preferences
							DDPref learnedDDPref = new DDPref();
							learnedDDPref.setData(evaluationData);
							learnedDDPref.setPreferences(learnedPrefs);
							Instances learnedSelectedSubset = learnedDDPref.wrapperSelectSubset(k);
							double learnedObjFnValue = learnedDDPref.getObjFnValue();
							System.out.println();
							System.out.println("Original Preferences:");
							System.out.println(prefs);
							System.out.println("Learned Preferences:");
							System.out.println(learnedPrefs);
							System.out.println("Learned Fn Value = "+learnedObjFnValue);
							System.out.println();
							ddpref.randomSelectSubset(k);
							System.out.println("True's Random =  "+ddpref.getObjFnValue());
							objFn_Random.addValue(ddpref.getObjFnValue());
							ddpref.topSelectSubset(k);
							System.out.println("True's TopK =    "+ddpref.getObjFnValue());
							objFn_TopK.addValue(ddpref.getObjFnValue());
							ddpref.greedySelectSubset(k);
							System.out.println("True's Greedy =  "+ddpref.getObjFnValue());
							objFn_Greedy.addValue(ddpref.getObjFnValue());
							ddpref.wrapperSelectSubset(k);
							System.out.println("True's Wrapper = "+ddpref.getObjFnValue());
							objFn_Wrapper.addValue(ddpref.getObjFnValue());
							System.out.println();
							double trueOnLearned = DDPref.computeObjFunction(learnedSelectedSubset, prefs);
							System.out.println("True Evaluating Learned's = "+trueOnLearned);
							objFn_Learned.addValue(trueOnLearned);
							System.out.println();
							System.out.println();
							/*
							time = System.currentTimeMillis();
							selectedSubset = ddpref.wrapperSelectSubset(k);
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_Wrapper += objFnValue;
							timeSum_Wrapper += time;
							if (verbose) System.out.println("Wrapper Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							//System.out.println("");
							
							
							time = System.currentTimeMillis();
							selectedSubset = ddpref.randomSelectSubset(k);
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_Random += objFnValue;
							timeSum_Random += time;
							if (verbose) System.out.println("Random Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							//System.out.println(selectedSubset.toString());
							
							time = System.currentTimeMillis();
							selectedSubset = ddpref.exhaustiveSelectSubset(k);
							time = System.currentTimeMillis() - time;
							objFnValue = ddpref.getObjFnValue();
							objFnSum_Exhaustive += objFnValue;
							timeSum_Exhaustive += time;
							if (verbose) System.out.println("Exhaustive Selected:  (objFn = " + objFnValue + ") (secs = "+time/1000+")");
							
							
							if (verbose) System.out.println();

							
							// generate the cross-product of the algorithms on the preferences
							if (trackCrossMeasure) {
								if (verbose) System.out.println("Cross-Measure for Greedy");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.greedySelectSubset);
								cmSum_Greedy = updateCMSum(crossMeasure,cmSum_Greedy);
								if (verbose) System.out.println(crossMeasure);
								if (verbose) System.out.println("Cross-Measure for Wrapper");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.wrapperSelectSubset);
								cmSum_Wrapper = updateCMSum(crossMeasure,cmSum_Wrapper);
								if (verbose) System.out.println(crossMeasure);
								if (verbose) System.out.println("Cross-Measure for Random");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.randomSelectSubset);
								cmSum_Random = updateCMSum(crossMeasure,cmSum_Random);
								if (verbose) System.out.println(crossMeasure);
								if (verbose) System.out.println("Cross-Measure for Exhaustive");
								crossMeasure=ddpref.crossMeasure(k,prefsList,SelectionMethod.exhaustiveSelectSubset);
								cmSum_Exhaustive = updateCMSum(crossMeasure,cmSum_Exhaustive);
								if (verbose) System.out.println(crossMeasure);
								if (verbose) System.out.println();
								
							}
							*/
						}
					
						System.out.println("For Alpha = "+alpha+" k = "+k+" prefs = "+prefs.name()+" over "+numTrials+" trials:");
						System.out.println("Summary info for Random:     objFn = "+objFn_Random.getMean()+" ("+objFn_Random.getStandardDeviation()+")");
						System.out.println("Summary info for TopK:       objFn = "+objFn_TopK.getMean()+" ("+objFn_TopK.getStandardDeviation()+")");
						System.out.println("Summary info for Greedy:     objFn = "+objFn_Greedy.getMean()+" ("+objFn_Greedy.getStandardDeviation()+")");
						System.out.println("Summary info for Wrapper:    objFn = "+objFn_Wrapper.getMean()+" ("+objFn_Wrapper.getStandardDeviation()+")");
						System.out.println("Summary info for Learned:    objFn = "+objFn_Learned.getMean()+" ("+objFn_Learned.getStandardDeviation()+")");
						
						System.out.println();
						/*
						trackCrossMeasure = false; // stop tracking cross measure after one run of trials
						*/
					}
				/*
					System.out.println("For Alpha = "+alpha+" k = "+k+" over "+numTrials+" trials:");
					System.out.println("Summary info for Cross-Measure Greedy over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Greedy,numTrials));
					System.out.println("Summary info for Cross-Measure Wrapper over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Wrapper,numTrials));
					System.out.println("Summary info for Cross-Measure Random over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Random,numTrials));
					System.out.println("Summary info for Cross-Measure Exhaustive over "+numTrials+" trials.");
					System.out.println(generateCMMeanOutput(crossMeasure,cmSum_Exhaustive,numTrials));
				*/
				}
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
		
	}
	
	

	

	
	/** Checks that a value is within the specified range.
	 * @param value The value to check.
	 * @param minValue The minimum value allowed.
	 * @param maxValue The maximum value allowed.
	 * @return <code>true</code> if minValue <= value <= maxValue, <code>false</code> otherwise.
	 */
	private static boolean valueInRange(double value, double minValue, double maxValue) {
		double EPSILON = 1.0E-5;
		return (value >= (minValue - EPSILON) && value <= (maxValue + EPSILON));
	}
	
	private static double[] parseDoubleArray(String strDoubles) {
		String[] strDoublesArray = strDoubles.split(",");
		double[] doubles = new double[strDoublesArray.length];
		for (int i=0; i<doubles.length; i++) {
			doubles[i] = Double.parseDouble(strDoublesArray[i]);
		}
		return doubles;
	}
	
	private static int[] parseIntegerArray(String strInts) {
		String[] strIntsArray = strInts.split(",");
		int[] ints = new int[strIntsArray.length];
		for (int i=0; i<ints.length; i++) {
			ints[i] = Integer.parseInt(strIntsArray[i]);
		}
		return ints;
	}
	
	private static DecimalFormat df = new DecimalFormat("0.0000");
	
	private static String arrayToString(double[] array) {
		String str = "";
		for (int i=0; i<array.length; i++) {
			str += df.format(array[i]);
			if (i != array.length-1) {
				str += ", ";
			}
		}
		return str;
	}

}
