/* DDPrefOptimization.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;

import java.text.DecimalFormat;

import weka.core.Optimization;

public class DDPrefOptimization extends Optimization {
	
	private double[][] m_meanQualities = null;
	private double[][] m_diversityMatches = null;
	private double[] m_objFnValues = null;
	
	protected boolean m_learningAlpha;
	protected double m_alpha;
	protected boolean m_verbose = false;
	
	public DDPrefOptimization(double[][] meanQualities, double[][] diversityMatches, double[] objFnValues) {
		super();
		//this.m_Debug = true;
		m_meanQualities = meanQualities;
		m_diversityMatches = diversityMatches;
		m_objFnValues = objFnValues;
		m_learningAlpha = true;
	}
	
	public DDPrefOptimization(double[][] meanQualities, double[][] diversityMatches, double[] objFnValues, double alpha) {
		this(meanQualities, diversityMatches, objFnValues);
		m_learningAlpha = false;
		m_alpha = alpha;
	}
	
	public void setVerbosity(boolean verbose) {
		m_verbose = verbose;
	}
	
	private double computeDDPref(int subsetIdx, double[] weights) {
		
		int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
		
		// compute the sum of weights
		double sumOfWeights = computeSumOfWeights(weights);
		
		// base case for all weights of 0
		if (sumOfWeights == 0) {
			return 0;
		}
		
		// retrieve the alpha parameter
		double alpha = (m_learningAlpha) ? weights[numAttributes] : m_alpha;
		
		// compute the sum of weighted qualities
		double qualityMatch = computeSumOfWeightedQualities(subsetIdx,weights);
		
		// compute the sum of the weighted diversity errors
		double diversityMatch = computeSumOfWeightedDiversities(subsetIdx,weights);
		
		// compute the DDPref Objective function for this subset
		double ddprefFn = ((1.0-alpha) * qualityMatch + alpha * diversityMatch) / sumOfWeights;
		
		return ddprefFn;	
	}
	
	private double computeSumOfWeights(double[] weights) {
		double sum = 0;
		int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
		for (int i=0; i<numAttributes; i++) {
			sum += weights[i];
		}
		return sum;
	}
	
	private double computeSumOfWeightedQualities(int subsetIdx, double[] weights) {
		
		int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
		
		// compute the sum of weighted qualities
		double qualityMatch = 0;
		for (int attIdx=0; attIdx<numAttributes; attIdx++) {
			qualityMatch += weights[attIdx] * m_meanQualities[subsetIdx][attIdx];
		}
		return qualityMatch;	
	}
	
	
	private double computeSumOfWeightedDiversities(int subsetIdx, double[] weights) {
		
		int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
		
		// compute the sum of the weighted diversity errors
		double diversityMatch = 0;
		for (int attIdx=0; attIdx<numAttributes; attIdx++) {
			diversityMatch += weights[attIdx] * m_diversityMatches[subsetIdx][attIdx];
		}
		return diversityMatch;	
	}
	

	
	
	/** Computes the objective function for a set of weights
	 * @param weights The weights of the objective function, alpha is the last weight.
	 * @return The value of the objective function.
	 */
	protected double objectiveFunction(double[] weights){
		
		double objFnValue = 0;
		
		// compute the sum of squared errors, with each set having a target value
		for (int subsetIdx=0; subsetIdx<m_meanQualities.length; subsetIdx++) {
			double ddprefFn = computeDDPref(subsetIdx, weights);
			// accumulate the squared error
			objFnValue += Math.pow(m_objFnValues[subsetIdx] - ddprefFn,2);
		}
		
		if (m_verbose) { 
			DecimalFormat df = new DecimalFormat("0.0000");
			System.out.print("Optimization Obj Fn = "+objFnValue+"  [ ");
			double maxWeight = Double.NEGATIVE_INFINITY;
			for (int i=0; i<weights.length; i++) {
				maxWeight = (weights[i]>maxWeight)?weights[i]:maxWeight;
			}
			for (int i=0; i<weights.length; i++)
				System.out.print(df.format(weights[i]/maxWeight)+" ");
			System.out.println("]");
		}
		
		if (Double.isNaN(objFnValue)) {
			System.err.println("Obj Fn is NaN!!!");
		}
		
		// return the squared error
		return objFnValue;
	}

	//TODO Incorporate computation of non-1 target objFnValues into gradient and hessian
	// Provide the first derivatives
	protected double[] evaluateGradient(double[] weights){
	    double[] gradient = new double[weights.length];
	    
	    int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
	    
	    // retrieve the alpha parameter
		double alpha = (m_learningAlpha) ? weights[numAttributes] : m_alpha;
	    
	    // compute the weightSum
		double weightSum = computeSumOfWeights(weights);
	    
	    // loop over each subset
	    for (int subsetIdx=0; subsetIdx<m_meanQualities.length; subsetIdx++) {
	    	
	    	double ddprefFn = computeDDPref(subsetIdx,weights);
			
			// compute the partial derivatives with respect to the attribute weights
			for (int attIdx=0; attIdx<numAttributes; attIdx++) {
				gradient[attIdx] += 
					2*(m_objFnValues[subsetIdx] - ddprefFn)  // changed 1.0 -> objFnValues
					* ((alpha*m_diversityMatches[subsetIdx][attIdx] - 
					     (1-alpha)*m_meanQualities[subsetIdx][attIdx]) / weightSum 
					   + ddprefFn / weightSum);
			}
			
			// compute the partial derivative with respect to alpha
			if (m_learningAlpha) {
				// compute the sum of weighted qualities
				double weightedQualitySum = computeSumOfWeightedQualities(subsetIdx,weights);
				// compute the sum of the weighted diversity errors
				double weightedDiversitySum = computeSumOfWeightedDiversities(subsetIdx,weights);
				// compute the gradient
				gradient[weights.length-1] += 
					-2 * (m_objFnValues[subsetIdx] - ddprefFn) // changed 1.0 -> objFnValues
					* (weightedDiversitySum - weightedQualitySum) / weightSum;
			}
	    }
	    
	    return gradient;
	}

	// If possible, provide the index^{th} row of the Hessian matrix
	protected double[] evaluateHessian(double[] weights, int index){
		
		double[] hessianRow = new double[weights.length];
		
		int numAttributes = (m_learningAlpha) ? weights.length-1 : weights.length;
	    
	    // retrieve the alpha parameter
		double alpha = (m_learningAlpha) ? weights[numAttributes] : m_alpha;
		
		// compute the weightSum
		double weightSum = computeSumOfWeights(weights);
		
		for (int subsetIdx=0; subsetIdx<m_meanQualities.length; subsetIdx++) {
			
			// precompute the parts
			// compute the ddpref fn
			double ddprefFn = computeDDPref(subsetIdx,weights);
			// compute the sum of weighted qualities
			double weightedQualitySum = computeSumOfWeightedQualities(subsetIdx,weights);
			// compute the sum of the weighted diversity errors
			double weightedDiversitySum = computeSumOfWeightedDiversities(subsetIdx,weights);
			
			// check whether we're doing the w_i rows, or the alpha row
			if (index >= 0 && index < numAttributes) { // the w_i row
			
				// the w_i columns, excluding alpha
				for (int index2=0; index2<numAttributes; index2++) {
					// second derivative, d / d w_i w_k
					hessianRow[index2] += 
						2 * ((ddprefFn 
						     - alpha*m_diversityMatches[subsetIdx][index2] 
						     - (1-alpha)*m_meanQualities[subsetIdx][index2]) / weightSum)
						  * ((ddprefFn 
						     - alpha*m_diversityMatches[subsetIdx][index] 
						     - (1-alpha)*m_meanQualities[subsetIdx][index]) / weightSum)
						+ 2 * (m_objFnValues[subsetIdx] - ddprefFn)  // changed 1.0 -> objFnValues
						    * (((1-alpha)*m_meanQualities[subsetIdx][index]
						        + alpha*m_diversityMatches[subsetIdx][index]
						        + (1-alpha)*m_meanQualities[subsetIdx][index2]
						        + alpha*m_diversityMatches[subsetIdx][index2]) 
						       / Math.pow(weightSum,2))
						       - (2 * ddprefFn / Math.pow(weightSum,2));
					                                                            
				}

				// processing alpha
				if (m_learningAlpha) {

					hessianRow[weights.length-1] += 
						   (-2 * ((ddprefFn 
							       - alpha*m_diversityMatches[subsetIdx][index] 
							       - (1-alpha)*m_meanQualities[subsetIdx][index]) 
							      / weightSum)
							   * (weightedDiversitySum - weightedQualitySum))
							/ weightSum
							+ 2 * (m_objFnValues[subsetIdx] - ddprefFn)    // changed 1.0 -> objFnValues
							    * ((weightedDiversitySum - weightedQualitySum)
							         / Math.pow(weightSum,2)
							       - ((m_diversityMatches[subsetIdx][index] 
							    	   - m_meanQualities[subsetIdx][index])
							    	  / weightSum));
							     
				}
			} 
			else if (index == weights.length-1) { // the alpha row
			
				// the w_i columns, excluding alpha
				for (int index2=0; index2<weights.length-1; index2++) {
				
					hessianRow[index2] += 
						   (-2 * ((ddprefFn 
							       - alpha*m_diversityMatches[subsetIdx][index2] 
							       - (1-alpha)*m_meanQualities[subsetIdx][index2]) 
							      / weightSum)
							   * (weightedDiversitySum - weightedQualitySum))
							/ weightSum
							- 2 * (m_objFnValues[subsetIdx] - ddprefFn)  // changed 1.0 -> objFnValues
							    * (m_diversityMatches[subsetIdx][index2] 
							       - m_meanQualities[subsetIdx][index2])
							    / weightSum
							+ 2 * (m_objFnValues[subsetIdx] - ddprefFn)  // changed 1.0 -> objFnValues
							    * (weightedDiversitySum - weightedQualitySum)
							    / Math.pow(weightSum,2);
				}
			
				// processing alpha
				hessianRow[hessianRow.length-1] += 
					2 * Math.pow(weightedDiversitySum - weightedQualitySum, 2)
					/ Math.pow(weightSum,2);
			}
			else {
				throw new IllegalArgumentException("Invalid index.  Must be between 0 and "+(weights.length-1)+", value is "+index+".");
			}
		}
		
		if (m_verbose) { System.out.println("Optimization Obj Fn = "+objectiveFunction(weights)); }

		return hessianRow;
	}
} 